include \masm32\include\masm32rt.inc
include \masm32\macros\timers.asm
.686

pbin	PROTO: DWORD, :DWORD

SizeTest	= 1
Use_dw2bin_ex	= 1
UsePbin	= 0

MinLampSize	= 200	; "default size" in case of external APIs, e.g. crt_strstr

LAMP$ MACRO cycles:REQ, csize:REQ	; Usage: print LAMP$(cycles, size), " LAMPs", 13, 10
	ffree ST(7)			;; free the stack for pushing
	push csize
	fild dword ptr [esp]		;; move into ST (0)
	pop eax
	fsqrt
	mov eax, cycles
	.if eax==0
		add eax, MinLampSize
	.endif
	push eax
	fild dword ptr [esp]		;; push cycles on FPU
	fmul ST, ST(1)		;; multiply with sqrt(size)
	fistp dword ptr [esp]
	pop eax
	invoke dwtoa, eax, addr LampsBuffer
EXITM <offset LampsBuffer>
ENDM

.data?
LampsBuffer	dd 4 dup (?)

Dw2BinBuffer	dd 9 dup (?)	; 32 are needed, we pad 4 for alignment

cy1	dd ?	; code speed
cy2	dd ?
cy3	dd ?
cy4	dd ?

cs1	dd ?	; code size
cs2	dd ?
cs3	dd ?
cs4	dd ?


.code

start:
	if 0
	print chr$("Testing correctness of results for Dword2Bin2", 13, 10)

	mov eax, 11110111000000001111111100000000b
	call Dword2Bin2
	print addr Dw2BinBuffer
	print chr$(13, 10, "11110111000000001111111100000000", 13, 10, 10)

	mov eax, 00001000111111110000000011111111b
	call Dword2Bin2
	print addr Dw2BinBuffer
	print chr$(13, 10, "00001000111111110000000011111111", 13, 10, 10)

	mov eax, 00100011111111000000001111111100b
	call Dword2Bin2
	print addr Dw2BinBuffer
	print chr$(13, 10, "00100011111111000000001111111100", 13, 10, 10)

	mov eax, 10101010101010101010101010101010b
	call Dword2Bin2
	print addr Dw2BinBuffer
	print chr$(13, 10, "10101010101010101010101010101010", 13, 10, 10)

	mov eax, 01010101010101010101010101010101b
	call Dword2Bin2
	print addr Dw2BinBuffer
	print chr$(13, 10, "01010101010101010101010101010101", 13, 10, 10, 10)
	endif

	NumLoops = 1000000
	invoke Sleep, 500

	if UsePbin
	LOOP_COUNT EQU NumLoops
	counter_begin LOOP_COUNT, HIGH_PRIORITY_CLASS
		invoke pbin, 11111000001111100000111111110000b, ADDR Dw2BinBuffer
		invoke pbin, 00001111100000111110000011111111b, ADDR Dw2BinBuffer
	counter_end
	mov cy4, eax

	else

	if SizeTest
		mov ecx, offset gL4
		sub ecx, offset gStart4	;pStart
		add ecx, 4
		add ecx, 2048	; bintable
		mov cs4, ecx
	endif
	mov eax, 60		; rough guess
	mov cy4, eax
	endif

	print str$(eax)," cycles timing pbin     ", 9, " "
	print str$(cs4), " bytes",9
	print LAMP$(cy4, cs4), " LAMPs", 13, 10

	if Use_dw2bin_ex
	LOOP_COUNT EQU NumLoops
	counter_begin LOOP_COUNT, HIGH_PRIORITY_CLASS
		invoke dw2bin_ex, 11111000001111100000111111110000b, ADDR Dw2BinBuffer
		invoke dw2bin_ex, 00001111100000111110000011111111b, ADDR Dw2BinBuffer
	counter_end
	mov cy1, eax
	mov cs1, 2140
	print str$(eax)," cycles timing dw2bin_ex", 9, " 2140 bytes", 9	; 92+2048
	print LAMP$(cy1, cs1), " LAMPs", 13, 10

	; int 3
	; invoke dw2bin_ex, 11111000001111100000111111110000b, ADDR Dw2BinBuffer
	endif

	LOOP_COUNT EQU NumLoops
	counter_begin LOOP_COUNT, HIGH_PRIORITY_CLASS
		mov eax, 11111000001111100000111111110000b
		call Dword2Bin
		mov eax, 00001111100000111110000011111111b
		call Dword2Bin
	counter_end
	mov cy2, eax
	print str$(eax)," cycles timing Dword2Bin", 9, " "
	print str$(cs2), " bytes",9
	print LAMP$(cy2, cs2), " LAMPs", 13, 10

	LOOP_COUNT EQU NumLoops
	counter_begin LOOP_COUNT, HIGH_PRIORITY_CLASS
		mov eax, 11111000001111100000111111110000b
		call Dword2Bin2
		mov eax, 00001111100000111110000011111111b
		call Dword2Bin2
	counter_end
	mov cy3, eax
	print str$(eax)," cycles timing Dword2Bin2", 9, " ",
	print str$(cs3), " bytes ",9
	print LAMP$(cy3, cs3), " LAMPs", 13, 10

	print chr$(13, 10), "LAMPs = Lean And Mean Points = cycles * sqrt(size)", 13,10

	getkey
	; .endif
	invoke ExitProcess, 0

align 4
Dword2Bin proc uses ebx
	if SizeTest
		mov ecx, offset gL2
		sub ecx, offset pStart
		add ecx, 4
		mov cs2, ecx
	endif

pStart:	mov edx, offset Dw2BinBuffer+32

	xor ecx, ecx
	mov [edx], cl	; null terminator
	add ecx, 31

@@:	xor ebx, ebx
	dec edx		; on exit, edx will point to the string
	sar eax, 1		; we work from right to left
	adc ebx, 48	; +48: Ascii 0 or, with carry, +49 : Ascii 1
	mov [edx], bl
	dec ecx		; decrement counter
	jge @B		; dec sets only the sign flag
	ret
Dword2Bin endp
gL2:

align 4
Dword2Bin2 proc uses ebx
	if SizeTest
		mov ecx, offset gL3
		sub ecx, offset pStart
		add ecx, 4
		mov cs3, ecx
	endif

pStart:	mov edx, offset Dw2BinBuffer
	xor ecx, ecx
	mov [edx+32], cl	; null terminator
	add ecx, 7	; 31

	; 4* setc bl plus add ebx, 30303030h is a lot slower
	; nop

@@:	sub ebx, ebx	; 2 cycle
	; ## this nop looks superfluous but leads to a decrease of up to 78 cycles ...!
	nop
	sar eax, 1		; 2 cycles - we read from left to right
	adc ebx, 48	; 1 +48: Ascii 0 or, with carry, +49 : Ascii 1

	sal ebx, 8		; 2 cycles - we write from right to left
	sar eax, 1		; 2 cycles - we read from left to right
	adc ebx, 48	; 1 +48: Ascii 0 or, with carry, +49 : Ascii 1

	sal ebx, 8		; 2 cycles - we write from right to left
	sar eax, 1		; 2 cycles - we read from left to right
	adc ebx, 48	; 1 +48: Ascii 0 or, with carry, +49 : Ascii 1

	sal ebx, 8		; 2 cycles - we write from right to left
	sar eax, 1		; 2 cycles - we read from left to right
	adc ebx, 48	; 1 +48: Ascii 0 or, with carry, +49 : Ascii 1

	mov [edx+4*ecx], ebx
	dec ecx		; 1 decrement counter
	; sub ecx, 1	; 1 but in practice 9 cycles slower
	jge @B		; dec sets only the sign flag
	ret
Dword2Bin2 endp
gL3:

align 16
OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE
gStart4:

pbin PROC PUBLIC num:DWORD,buf:DWORD
pStart:
    push ebx
    push esi
    push edi
   
    mov ecx,16[esp]
    mov edx,20[esp]

	if UsePbin		; for commenting out with a switch
    mov edi, offset bintable		; where defined??
   	endif

    movzx eax,cl
    mov ebx,[edi+eax*8]
    mov esi,[edi+eax*8+4]
    mov 24[edx],ebx
    mov 28[edx],esi

    movzx eax,ch
    mov ebx,[edi+eax*8]
    mov esi,[edi+eax*8+4]
    shr ecx,16
    mov 16[edx],ebx
    mov 20[edx],esi

    movzx eax,cl
    mov ebx,[edi+eax*8]
    mov esi,[edi+eax*8+4]
    mov 8[edx],ebx
    mov 12[edx],esi

    movzx eax,ch
    mov ebx,[edi+eax*8]
    mov esi,[edi+eax*8+4]
    sub eax,eax
    mov [edx],ebx
    mov 4[edx],esi

    mov 32[edx],al

    pop edi
    pop esi
    pop ebx
    ret 8
pbin ENDP
gL4:

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef 

end start

#########  Disassembly dw2bin_ex  #########

92 bytes plus 2048 bytes table

CPU Disasm
Address    Hex dump          Command                            Comments
00401340   $ 55             push ebp
00401341   . 8BEC           mov ebp, esp
00401343   . 53             push ebx
00401344   . BB B0304000    mov ebx, offset TestBits.004030B0  ; ASCII "000000000000000100000010000000110000010000000101000001100000011100001000000010010000101000001011000011000000110100001110000011110001000000010001000100100001001100010100000101010001011000010111000110000001100100011010000110110001110000011"...
00401349   . 33C0           xor eax, eax
0040134B   . 8B55 0C        mov edx, dword ptr [arg.2]         ; ASCII "00001111100000111110000011111111"
0040134E   . 8A45 0B        mov al, byte ptr [arg.1+3]
00401351   . 8B0CC3         mov ecx, dword ptr [eax*8+ebx]
00401354   . 890A           mov dword ptr [edx], ecx
00401356   . 8B4CC3 04      mov ecx, dword ptr [eax*8+ebx+4]
0040135A   . 894A 04        mov dword ptr [edx+4], ecx
0040135D   . 8A45 0A        mov al, byte ptr [arg.1+2]
00401360   . 8B0CC3         mov ecx, dword ptr [eax*8+ebx]
00401363   . 894A 08        mov dword ptr [edx+8], ecx
00401366   . 8B4CC3 04      mov ecx, dword ptr [eax*8+ebx+4]
0040136A   . 894A 0C        mov dword ptr [edx+0C], ecx
0040136D   . 8A45 09        mov al, byte ptr [arg.1+1]
00401370   . 8B0CC3         mov ecx, dword ptr [eax*8+ebx]
00401373   . 894A 10        mov dword ptr [edx+10], ecx
00401376   . 8B4CC3 04      mov ecx, dword ptr [eax*8+ebx+4]
0040137A   . 894A 14        mov dword ptr [edx+14], ecx
0040137D   . 8A45 08        mov al, byte ptr [arg.1]
00401380   . 8B0CC3         mov ecx, dword ptr [eax*8+ebx]
00401383   . 894A 18        mov dword ptr [edx+18], ecx
00401386   . 8B4CC3 04      mov ecx, dword ptr [eax*8+ebx+4]
0040138A   . 894A 1C        mov dword ptr [edx+1C], ecx
0040138D   . C642 20 00     mov byte ptr [edx+20], 0
00401391   . 5B             pop ebx
00401392   . C9             leave
00401393   . C2 0800        retn 8
00401396-00401340=92

LeanAndMeanPoints:
139 cycles timing dw2bin_ex	size 2140 bytes	139*sqrt(2140)=6384
462 cycles timing Dword2Bin	size 31 bytes	462*sqrt(31)=2572
335 cycles timing Dword2Bin2	size 57 bytes	335*sqrt(57)=2529

OPT_DelTmp	0
